/*
  histograms.cc

  Written by Taiji Yamada <taiji@aihara.co.jp>
*/
#include <iostream>
#include <vector>
#include <list>
#include <functional>
#if __cplusplus < 201103L
#include <boost/function.hpp>
#endif
#include <cstdlib>      // random(), etc
#include <cassert>
#if __cplusplus < 201103L
#include <boost/utility.hpp>
#endif
#include "statistics.hpp"
#include "histograms.hpp"
#include "histograms_online.hpp"

#define numberof(a) (sizeof(a)/sizeof(a[0]))

//#define USE_LOGARITHMIC_HISTOGRAM
//#define USE_LIST_FOR_MOVING_STORAGE

int main()
{
  const double maximum_permissible_error(1e-3);
  const char *result_names[] = {
    "minimum value",
    "maximum value",
    "histogram min_element",
    "histogram min_value",
    "histogram max_element",
    "histogram max_value",
    "histogram mean",
    "histogram sigma",
    "histogram sum",
  };
  std::vector<double> v(10000);
  const size_t n_bins(v.size()/10), n_prior_points(v.size()/2);
#if 0 /* WITHOUT_HISTOGRAM_ONLINE */
#else
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
  histogram_online<double> histogram_online(n_bins, n_prior_points);
#else
  logarithmic_histogram_online<double> histogram_online(n_bins, n_prior_points);
#endif
#endif
  srandomdev();
  for (size_t i=0; i<v.size(); ++i) {
    double x((double)random()/((1L<<31)-1));
    v[i] = x;
#if 0 /* WITHOUT_HISTOGRAM_ONLINE */
#else
    histogram_online.add(x);
#endif
  }
  double results[numberof(result_names)] = {
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
    *min_element(v.begin(), v.end()),
#else
    positive_min_value(v.begin(), v.end()),
#endif
    *max_element(v.begin(), v.end()),
  };
  std::vector<size_t> histogram_result(n_bins);
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
  size_t histogram_result_summation(histogram(histogram_result, v.begin(), v.end(), results[0], results[1]));
#else
  size_t histogram_result_summation(histogram_logarithm(histogram_result, v.begin(), v.end(), results[0], results[1]));
#endif
  {
    double more_results[] = {
      (double)std::distance(histogram_result.begin(), min_element(histogram_result.begin(), histogram_result.end())),
      (double)*min_element(histogram_result.begin(), histogram_result.end()),
      (double)std::distance(histogram_result.begin(), max_element(histogram_result.begin(), histogram_result.end())),
      (double)*max_element(histogram_result.begin(), histogram_result.end()),
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
      histogram_mean(histogram_result, results[0], results[1]),
      histogram_standard_deviation(histogram_result, results[0], results[1]),
#else
      histogram_logarithm_mean(histogram_result, results[0], results[1]),
      histogram_logarithm_standard_deviation(histogram_result, results[0], results[1]),
#endif
#if 0
      (double)std::accumulate(histogram_result.begin(), histogram_result.end(), 0),
#else
      (double)histogram_result_summation,
#endif
    };
    std::copy(more_results, more_results + numberof(more_results), &results[2]);
  }
#if 0 /* WITHOUT_HISTOGRAM_ONLINE */
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
  double continue_result_range[] = { *min_element(v.begin(), v.begin() + n_prior_points), *max_element(v.begin(), v.begin() + n_prior_points), };
#else
  double continue_result_range[] = { positive_min_value(v.begin(), v.begin() + n_prior_points), *max_element(v.begin(), v.begin() + n_prior_points), };
#endif
  double continue_results[numberof(result_names)] = {
  };
  std::vector<size_t> histogram_continue_result(n_bins);
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
  size_t histogram_continue_result_summation(histogram(histogram_continue_result, v.begin(), v.begin() + n_prior_points, continue_result_range[0], continue_result_range[1]));
  histogram_continue_result_summation += histogram(histogram_continue_result, v.begin() + n_prior_points, v.end(), continue_result_range[0], continue_result_range[1]);
#else
  size_t histogram_continue_result_summation(histogram_logarithm(histogram_continue_result, v.begin(), v.begin() + n_prior_points, continue_result_range[0], continue_result_range[1]));
  histogram_continue_result_summation += histogram_logarithm(histogram_continue_result, v.begin() + n_prior_points, v.end(), continue_result_range[0], continue_result_range[1]);
#endif
  {
    double more_results[] = {
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
      *min_element(v.begin(), v.end()),
#else
      positive_min_value(v.begin(), v.end()),
#endif
      *max_element(v.begin(), v.end()),
      (double)std::distance(histogram_continue_result.begin(), min_element(histogram_continue_result.begin(), histogram_continue_result.end())),
      (double)*min_element(histogram_continue_result.begin(), histogram_continue_result.end()),
      (double)std::distance(histogram_continue_result.begin(), max_element(histogram_continue_result.begin(), histogram_continue_result.end())),
      (double)*max_element(histogram_continue_result.begin(), histogram_continue_result.end()),
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
      histogram_mean(histogram_continue_result, continue_result_range[0], continue_result_range[1]),
      histogram_standard_deviation(histogram_continue_result, continue_result_range[0], continue_result_range[1]),
#else
      histogram_logarithm_mean(histogram_continue_result, continue_result_range[0], continue_result_range[1]),
      histogram_logarithm_standard_deviation(histogram_continue_result, continue_result_range[0], continue_result_range[1]),
#endif
#if 0
      (double)std::accumulate(histogram_continue_result.begin(), histogram_continue_result.end(), 0),
#else
      (double)histogram_continue_result_summation,
#endif
    };
    std::copy(more_results, more_results + numberof(more_results), &continue_results[0]);
  }
  for (size_t j=0; j<numberof(continue_results); ++j) {
    double error(sqrt(pow(continue_results[j] - results[j], 2)) / (results[j] ? fabs(results[j]) : 1));
    std::cout << result_names[j] << ":\t" << continue_results[j] << "\t" << results[j] << "\t" << error << std::endl;
    if (!(j==2||j==3||j==4||j==5)) assert(error < maximum_permissible_error);
  }
#else
  double online_results[numberof(result_names)] = {
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
    histogram_online.get_minimum_value(),
#else
    histogram_online.get_positive_minimum_value(),
#endif
    histogram_online.get_maximum_value(),
#if !0
    (double)histogram_online.get_minimum_height_index(),
    (double)histogram_online.get_minimum_height(),
    (double)histogram_online.get_maximum_height_index(),
    (double)histogram_online.get_maximum_height(),
    histogram_online.get_heights_mean(),
    histogram_online.get_heights_standard_deviation(),
    (double)histogram_online.get_heights_summation(),
#else
    (double)std::distance(histogram_online.get_heights().begin(), min_element(histogram_online.get_heights().begin(), histogram_online.get_heights().end())),
    (double)*min_element(histogram_online.get_heights().begin(), histogram_online.get_heights().end()),
    (double)std::distance(histogram_online.get_heights().begin(), max_element(histogram_online.get_heights().begin(), histogram_online.get_heights().end())),
    (double)*max_element(histogram_online.get_heights().begin(), histogram_online.get_heights().end()),
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
    histogram_mean(histogram_online.get_heights(), histogram_online.get_lower_bound(0), histogram_online.get_upper_bound(n_bins-1)),
    histogram_standard_deviation(histogram_online.get_heights(), histogram_online.get_lower_bound(0), histogram_online.get_upper_bound(n_bins-1)),
#else
    histogram_logarithm_mean(histogram_online.get_heights(), histogram_online.get_lower_bound(0), histogram_online.get_upper_bound(n_bins-1)),
    histogram_logarithm_standard_deviation(histogram_online.get_heights(), histogram_online.get_lower_bound(0), histogram_online.get_upper_bound(n_bins-1)),
#endif
    (double)std::accumulate(histogram_online.get_heights().begin(), histogram_online.get_heights().end(), 0),
#endif
  };
  for (size_t j=0; j<numberof(online_results); ++j) {
    double error(sqrt(pow(online_results[j] - results[j], 2)) / (results[j] ? fabs(results[j]) : 1));
    std::cout << result_names[j] << ":\t" << online_results[j] << "\t" << results[j] << "\t" << error << std::endl;
    if (!(j==2||j==3||j==4||j==5)) assert(error < maximum_permissible_error);
  }
#endif

#if !defined(USE_LIST_FOR_MOVING_STORAGE)
  std::vector<double> m(v.size());
#else
  std::list<double> m;
#endif
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
  moving_histogram_online<double> moving_histogram_online(n_bins, v.size());
#else
  moving_logarithmic_histogram_online<double> moving_histogram_online(n_bins, v.size());
#endif
  for (size_t i=0; i<v.size()*10; ++i) {
    double x((double)random()/((1L<<31)-1));
#if !defined(USE_LIST_FOR_MOVING_STORAGE)
    m[i % v.size()] = x;
#else
    if (!(i < v.size())) m.pop_front();
    m.push_back(x);
#endif
    moving_histogram_online.add(x);
    if (i < v.size()) continue;
    double moving_results[numberof(result_names)] = {
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
      *min_element(m.begin(), m.end()),
#else
      positive_min_value(m.begin(), m.end()),
#endif
      *max_element(m.begin(), m.end()),
    };
    std::vector<size_t> moving_histogram_result(n_bins);
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
    size_t moving_histogram_result_summation(histogram(moving_histogram_result, m.begin(), m.end(), moving_results[0], moving_results[1]));
#else
    size_t moving_histogram_result_summation(histogram_logarithm(moving_histogram_result, m.begin(), m.end(), moving_results[0], moving_results[1]));
#endif
    {
      double more_results[] = {
	(double)std::distance(moving_histogram_result.begin(), min_element(moving_histogram_result.begin(), moving_histogram_result.end())),
	(double)*min_element(moving_histogram_result.begin(), moving_histogram_result.end()),
	(double)std::distance(moving_histogram_result.begin(), max_element(moving_histogram_result.begin(), moving_histogram_result.end())),
	(double)*max_element(moving_histogram_result.begin(), moving_histogram_result.end()),
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
	histogram_mean(moving_histogram_result, moving_results[0], moving_results[1]),
	histogram_standard_deviation(moving_histogram_result, moving_results[0], moving_results[1]),
#else
	histogram_logarithm_mean(moving_histogram_result, moving_results[0], moving_results[1]),
	histogram_logarithm_standard_deviation(moving_histogram_result, moving_results[0], moving_results[1]),
#endif
#if 0
	(double)std::accumulate(moving_histogram_result.begin(), moving_histogram_result.end(), 0),
#else
	(double)moving_histogram_result_summation,
#endif
      };
      std::copy(more_results, more_results + numberof(more_results), &moving_results[2]);
    }
    double online_moving_results[numberof(result_names)] = {
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
      moving_histogram_online.get_minimum_value(),
#else
      moving_histogram_online.get_positive_minimum_value(),
#endif
      moving_histogram_online.get_maximum_value(),
#if !0
      (double)moving_histogram_online.get_minimum_height_index(),
      (double)moving_histogram_online.get_minimum_height(),
      (double)moving_histogram_online.get_maximum_height_index(),
      (double)moving_histogram_online.get_maximum_height(),
      moving_histogram_online.get_heights_mean(),
      moving_histogram_online.get_heights_standard_deviation(),
      (double)moving_histogram_online.get_heights_summation(),
#else
      (double)std::distance(moving_histogram_online.get_heights().begin(), min_element(moving_histogram_online.get_heights().begin(), moving_histogram_online.get_heights().end())),
      (double)*min_element(moving_histogram_online.get_heights().begin(), moving_histogram_online.get_heights().end()),
      (double)std::distance(moving_histogram_online.get_heights().begin(), max_element(moving_histogram_online.get_heights().begin(), moving_histogram_online.get_heights().end())),
      (double)*max_element(moving_histogram_online.get_heights().begin(), moving_histogram_online.get_heights().end()),
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
      histogram_mean(moving_histogram_online.get_heights(), moving_histogram_online.get_lower_bound(0), moving_histogram_online.get_upper_bound(n_bins-1)),
      histogram_standard_deviation(moving_histogram_online.get_heights(), moving_histogram_online.get_lower_bound(0), moving_histogram_online.get_upper_bound(n_bins-1)),
#else
      histogram_logarithm_mean(moving_histogram_online.get_heights(), moving_histogram_online.get_lower_bound(0), moving_histogram_online.get_upper_bound(n_bins-1)),
      histogram_logarithm_standard_deviation(moving_histogram_online.get_heights(), moving_histogram_online.get_lower_bound(0), moving_histogram_online.get_upper_bound(n_bins-1)),
#endif
      (double)std::accumulate(moving_histogram_online.get_heights().begin(), moving_histogram_online.get_heights().end(), 0),
#endif
    };
    for (size_t j=0; j<numberof(online_moving_results); ++j) {
      double error(sqrt(pow(online_moving_results[j] - moving_results[j], 2)) / (moving_results[j] ? fabs(moving_results[j]) : 1));
      if (i+1 == v.size()*10)
        std::cout << "moving " << result_names[j] << ":\t" << online_moving_results[j] << "\t" << moving_results[j] << "\t" << error << std::endl;
      if (!(j==2||j==3||j==4||j==5)) assert(error < maximum_permissible_error);
    }
  }
  return 0;
}
