/*
  histograms_online.hpp

  Copyright (C) 2016 Taiji Yamada <taiji@aihara.co.jp>
*/
#ifndef _HISTOGRAMS_ONLINE_HPP_
#define _HISTOGRAMS_ONLINE_HPP_
#include "statistics_online.hpp"
#include "histograms.hpp"

template <typename T>
class histogram_online : public statistics_online<T> {
public:
  histogram_online(const size_t n_bins, const T lb, const T ub, const bool option_lt_ub = false) : statistics_online<T>(),
    heights(n_bins), lb(lb), ub(ub), option_lt_ub(option_lt_ub), n_prior_points(0), summation(0), prior_points(0) {}
  histogram_online(const size_t n_bins, const size_t n_prior_points, const bool option_lt_ub = false) : statistics_online<T>(),
    heights(n_bins), lb(0), ub(0), option_lt_ub(option_lt_ub), n_prior_points(n_prior_points), summation(0)
  {
    if (1 < n_prior_points)
      prior_points = new T[n_prior_points];
  }
  virtual ~histogram_online()
  {
    if (prior_points)
      delete [] prior_points;
    prior_points = 0;
  }
  void add(const T v)
  {
    const size_t count(statistics_online<T>::get_count());
    statistics_online<T>::add(v);
    if (1 < n_prior_points) {
      if (count < n_prior_points) {
	prior_points[count] = v;
	if (count + 1 == n_prior_points) {
	  lb = *std::min_element(prior_points, prior_points + n_prior_points);
	  ub = *std::max_element(prior_points, prior_points + n_prior_points);
	  summation = histogram(heights, prior_points, prior_points + n_prior_points, lb, ub);
	}
      }
      else
	summation += histogram(heights, &v, &v + 1, lb, ub);
    }
    else
      if (!(lb == ub))
	summation += histogram(heights, &v, &v + 1, lb, ub);
  }
  void reset() { statistics_online<T>::reset(); fill(heights.begin(), heights.end(), 0); summation = 0; }

  const std::vector<size_t> &get_heights() const
  {
    return heights;
  }
  size_t get_height(const size_t i) const
  {
    return heights[i];
  }
  T get_lower_bound(const size_t i) const
  {
    return histogram_lower_bound(i, heights.size(), lb, ub, option_lt_ub);
  }
  T get_upper_bound(const size_t i) const
  {
    return histogram_upper_bound(i, heights.size(), lb, ub, option_lt_ub);
  }

  size_t get_minimum_height_index() const
  {
    return std::distance(heights.begin(), min_element(heights.begin(), heights.end()));
  }
  T get_minimum_height() const
  {
    return *min_element(heights.begin(), heights.end());
  }
  size_t get_maximum_height_index() const
  {
    return std::distance(heights.begin(), max_element(heights.begin(), heights.end()));
  }
  T get_maximum_height() const
  {
    return *max_element(heights.begin(), heights.end());
  }

  T get_heights_mean() const
  {
    return histogram_mean<T>(heights, lb, ub);
  }
  T get_heights_variance() const
  {
    return histogram_variance<T>(heights, lb, ub);
  }
  T get_heights_standard_deviation() const
  {
    return histogram_standard_deviation<T>(heights, lb, ub);
  }
  size_t get_heights_summation() const
  {
    return summation;
  }
protected:
  std::vector<size_t> heights;
  T lb, ub;
  bool option_lt_ub;
  const size_t n_prior_points;
  size_t summation;
  T *prior_points;
};

template <typename T>
class logarithmic_histogram_online : public histogram_online<T> {
public:
  logarithmic_histogram_online(const size_t n_bins, const T lb, const T ub, const bool option_lt_ub = false) : histogram_online<T>(n_bins, lb, ub, option_lt_ub) {}
  logarithmic_histogram_online(const size_t n_bins, const size_t n_prior_points, const bool option_lt_ub = false) : histogram_online<T>(n_bins, n_prior_points, option_lt_ub) {}
  void add(const T v)
  {
    const size_t count(statistics_online<T>::get_count());
    std::vector<size_t> &heights(histogram_online<T>::heights);
    T &lb(histogram_online<T>::lb), &ub(histogram_online<T>::ub);
    const size_t n_prior_points(histogram_online<T>::n_prior_points);
    size_t &summation(histogram_online<T>::summation);
    T *prior_points(histogram_online<T>::prior_points);
    statistics_online<T>::add(v);
    if (1 < n_prior_points) {
      if (count < n_prior_points) {
	prior_points[count] = v;
	if (count + 1 == n_prior_points) {
	  lb = positive_min_value(prior_points, prior_points + n_prior_points);
	  ub = *std::max_element(prior_points, prior_points + n_prior_points);
	  summation = histogram_logarithm(heights, prior_points, prior_points + n_prior_points, lb, ub);
	}
      }
      else
	summation += histogram_logarithm(heights, &v, &v + 1, lb, ub);
    }
    else
      if (!(lb == ub))
	summation += histogram_logarithm(heights, &v, &v + 1, lb, ub);
  }

  T get_lower_bound(const size_t i) const
  {
    const std::vector<size_t> &heights(histogram_online<T>::heights);
    const T lb(histogram_online<T>::lb), ub(histogram_online<T>::ub);
    const bool option_lt_ub(histogram_online<T>::option_lt_ub);

    return histogram_logarithm_lower_bound(i, heights.size(), lb, ub, option_lt_ub);
  }
  T get_upper_bound(const size_t i) const
  {
    const std::vector<size_t> &heights(histogram_online<T>::heights);
    const T lb(histogram_online<T>::lb), ub(histogram_online<T>::ub);
    const bool option_lt_ub(histogram_online<T>::option_lt_ub);

    return histogram_logarithm_upper_bound(i, heights.size(), lb, ub, option_lt_ub);
  }

  T get_heights_mean() const
  {
    const std::vector<size_t> &heights(histogram_online<T>::heights);
    const T lb(histogram_online<T>::lb), ub(histogram_online<T>::ub);

    return histogram_logarithm_mean<T>(heights, lb, ub);
  }
  T get_heights_variance() const
  {
    const std::vector<size_t> &heights(histogram_online<T>::heights);
    const T lb(histogram_online<T>::lb), ub(histogram_online<T>::ub);

    return histogram_logarithm_variance<T>(heights, lb, ub);
  }
  T get_heights_standard_deviation() const
  {
    const std::vector<size_t> &heights(histogram_online<T>::heights);
    const T lb(histogram_online<T>::lb), ub(histogram_online<T>::ub);

    return histogram_logarithm_standard_deviation<T>(heights, lb, ub);
  }
};

template <typename T>
class moving_histogram_online : public moving_statistics_online<T> {
public:
  moving_histogram_online(const size_t n_bins, const T lb, const T ub, const size_t size = 100, const bool option_lt_ub = false) : moving_statistics_online<T>(size), heights(n_bins), lb(lb), ub(ub), option_lt_ub(option_lt_ub), summation(0) {}
  moving_histogram_online(const size_t n_bins, const size_t size = 100, const bool option_lt_ub = false) : moving_statistics_online<T>(size), heights(n_bins), lb(0), ub(0), option_lt_ub(option_lt_ub), summation(0) {}
  void add(const typename moving_statistics_online<T>::value_type v)
  {
    const size_t size(moving_statistics_online<T>::size);
    const size_t index(moving_statistics_online<T>::get_index());
    const T *storage(moving_statistics_online<T>::get_storage());
    const typename moving_statistics_online<T>::value_type ov(storage[index]);
    const bool previous_initial(moving_statistics_online<T>::is_initial());
    moving_statistics_online<T>::add(v);
    const bool initial(moving_statistics_online<T>::is_initial());
    if (!initial) {
      if (previous_initial) {
	if (lb == ub) {
	  lb = *std::min_element(storage, storage + size);
	  ub = *std::max_element(storage, storage + size);
	}
	summation = histogram(heights, storage, storage + size, lb, ub);
      }
      else {
	bool update(false);
	if (ov < lb || v < lb) { lb = *std::min_element(storage, storage + size); update = true; }
	if (ub < ov || ub < v) { ub = *std::max_element(storage, storage + size); update = true; }
	if (update) {
	  fill(heights.begin(), heights.end(), 0);
	  summation = histogram(heights, storage, storage + size, lb, ub);
	}
	else {
	  const typename moving_statistics_online<T>::value_type sz(ub - lb);
	  const size_t oid((ov - lb)/sz*heights.size()), id((v - lb)/sz*heights.size());
	  if (oid < heights.size())
	    --heights[oid];
	  if (id < heights.size())
	    ++heights[id];
	}
      }
    }
  }
  void reset() { moving_statistics_online<T>::reset(); fill(heights.begin(), heights.end(), 0); summation = 0; }

  const std::vector<size_t> &get_heights() const
  {
    return heights;
  }
  size_t get_height(const size_t i) const
  {
    return heights[i];
  }
  T get_lower_bound(const size_t i) const
  {
    return histogram_lower_bound(i, heights.size(), lb, ub, option_lt_ub);
  }
  T get_upper_bound(const size_t i) const
  {
    return histogram_upper_bound(i, heights.size(), lb, ub, option_lt_ub);
  }

  size_t get_minimum_height_index() const
  {
    return std::distance(heights.begin(), min_element(heights.begin(), heights.end()));
  }
  T get_minimum_height() const
  {
    return *min_element(heights.begin(), heights.end());
  }
  size_t get_maximum_height_index() const
  {
    return std::distance(heights.begin(), max_element(heights.begin(), heights.end()));
  }
  T get_maximum_height() const
  {
    return *max_element(heights.begin(), heights.end());
  }

  T get_heights_mean() const
  {
    return histogram_mean<T>(heights, lb, ub);
  }
  T get_heights_variance() const
  {
    return histogram_variance<T>(heights, lb, ub);
  }
  T get_heights_standard_deviation() const
  {
    return histogram_standard_deviation<T>(heights, lb, ub);
  }
  size_t get_heights_summation() const
  {
    return summation;
  }
protected:
  std::vector<size_t> heights;
  T lb, ub;
  bool option_lt_ub;
  size_t summation;
};

template <typename T>
class moving_logarithmic_histogram_online : public moving_histogram_online<T> {
public:
  moving_logarithmic_histogram_online(const size_t n_bins, const T lb, const T ub, const size_t size = 100, const bool option_lt_ub = false) : moving_histogram_online<T>(n_bins, lb, ub, size, option_lt_ub) {}
  moving_logarithmic_histogram_online(const size_t n_bins, const size_t size = 100, const bool option_lt_ub = false) : moving_histogram_online<T>(n_bins, size, option_lt_ub) {}
  void add(const typename moving_histogram_online<T>::value_type v)
  {
    const size_t size(moving_histogram_online<T>::size);
    const size_t index(moving_histogram_online<T>::get_index());
    const T *storage(moving_histogram_online<T>::get_storage());
    const typename moving_histogram_online<T>::value_type ov(storage[index]);
    const bool previous_initial(moving_histogram_online<T>::is_initial());
    moving_statistics_online<T>::add(v);
    const bool initial(moving_histogram_online<T>::is_initial());
    std::vector<size_t> &heights(moving_histogram_online<T>::heights);
    T &lb(moving_histogram_online<T>::lb), &ub(moving_histogram_online<T>::ub);
    size_t &summation(moving_histogram_online<T>::summation);
    if (!initial) {
      if (previous_initial) {
	if (lb == ub) {
	  lb = positive_min_value(storage, storage + size);
	  ub = *std::max_element(storage, storage + size);
	}
	summation = histogram_logarithm(heights, storage, storage + size, lb, ub);
      }
      else {
	bool update(false);
	if (ov < lb || v < lb) { lb = positive_min_value(storage, storage + size); update = true; }
	if (ub < ov || ub < v) { ub = *std::max_element(storage, storage + size); update = true; }
	if (update) {
	  fill(heights.begin(), heights.end(), 0);
	  summation = histogram_logarithm(heights, storage, storage + size, lb, ub);
	}
	else {
	  const typename moving_histogram_online<T>::value_type sz(log(ub) - log(lb));
	  const size_t oid((log(ov) - log(lb))/sz*heights.size()), id((log(v) - log(lb))/sz*heights.size());
	  if (oid < heights.size())
	    --heights[oid];
	  if (id < heights.size())
	    ++heights[id];
	}
      }
    }
  }

  T get_lower_bound(const size_t i) const
  {
    const std::vector<size_t> &heights(moving_histogram_online<T>::heights);
    const T lb(moving_histogram_online<T>::lb), ub(moving_histogram_online<T>::ub);
    const bool option_lt_ub(moving_histogram_online<T>::option_lt_ub);

    return histogram_logarithm_lower_bound(i, heights.size(), lb, ub, option_lt_ub);
  }
  T get_upper_bound(const size_t i) const
  {
    const std::vector<size_t> &heights(moving_histogram_online<T>::heights);
    const T lb(moving_histogram_online<T>::lb), ub(moving_histogram_online<T>::ub);
    const bool option_lt_ub(moving_histogram_online<T>::option_lt_ub);

    return histogram_logarithm_upper_bound(i, heights.size(), lb, ub, option_lt_ub);
  }

  T get_heights_mean() const
  {
    const std::vector<size_t> &heights(moving_histogram_online<T>::heights);
    const T lb(moving_histogram_online<T>::lb), ub(moving_histogram_online<T>::ub);

    return histogram_logarithm_mean<T>(heights, lb, ub);
  }
  T get_heights_variance() const
  {
    const std::vector<size_t> &heights(moving_histogram_online<T>::heights);
    const T lb(moving_histogram_online<T>::lb), ub(moving_histogram_online<T>::ub);

    return histogram_logarithm_variance<T>(heights, lb, ub);
  }
  T get_heights_standard_deviation() const
  {
    const std::vector<size_t> &heights(moving_histogram_online<T>::heights);
    const T lb(moving_histogram_online<T>::lb), ub(moving_histogram_online<T>::ub);

    return histogram_logarithm_standard_deviation<T>(heights, lb, ub);
  }
};

#endif
